/***************************************************************************
 *   Copyright (C) 2007 by Antony Martin                                   *
 *   martin.antony@yahoo.fr                                                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* Cree le : 5 janvier 2007
   derniere modification le 13/03/2007 */


/*
    MediaManager
    Interface
    
    Chargement de medias a partir de fichiers
    et enregistrement des fonctions d'importation/exportation.
*/

/*
 * 15/03/07 --> renommage des fonctions, prefixe MediaManager change en Media
 */


#include <SCE/SCEMinimal.h>

#include <SCE/utils/SCEString.h>

#include <SCE/interface/SCEMedia.h>


/* structure representant la fonction a utiliser pour le chargement d'un media definie */
typedef struct sce_smedia_funcload SCE_SFuncLoad;
struct sce_smedia_funcload
{
    /* type du media que la fonction charge
       identifiant utilise en interne */
    int type_id;
    
    /* nombre magic du fichier */
    int magic;
    
    /* extension(s) du fichier */
    char *ext;
    
    /* fonction de chargement */
    /**
      * \param FILE* : structure du fichier ouvert en lecture seule
      * \param const-char* : nom du fichier
      * \param void* : donnees utilisateur
      */
    void* (*load)(FILE*, const char*, void*);
};

typedef struct sce_smedia_funcsave SCE_SFuncSave;
struct sce_smedia_funcsave
{
    /* type du media que la fonction charge
       identifiant utilise en interne */
    int type_id;
    
    /* fonction de sauvegarde */
    /**
      * \param FILE* : structure du fichier ouvert en ecriture  (...)
      * \param const-char* : nom du fichier
      * \param void* : donnees utilisateur (en principe le media a sauvegarder)
      */
    int (*save)(/*FILE*, */const char*, void*);
};


/* nombre de fonctions de chargement */
static int n_lFuncs;

/* fonctions de chargement de medias (Load Funcs) */
static SCE_SFuncLoad *l_Func;


/* nombre de fonctions de sauvegarde */
static int n_sFuncs;

/* fonctions de sauvegarde de medias (Save Funcs) */
static SCE_SFuncSave *s_Func;


int SCE_Media_Init(void)
{
    n_lFuncs = 0;
    l_Func = NULL;
    
    n_sFuncs = 0;
    s_Func = NULL;
    
    return (SCE_OK); /* ... */
}


int SCE_Media_Quit(void)
{
    int code = SCE_OK;
    
    if(l_Func && SCE_free(l_Func) < 0)
    {
        Logger_LogSrc();
        code--;
    }
    l_Func = NULL;
    n_lFuncs = 0;
    
    if(s_Func && SCE_free(s_Func) < 0)
    {
        Logger_LogSrc();
        /* si deux erreurs consecutives, code = SCE_OK - 2 */
        code--;
    }
    s_Func = NULL;
    n_sFuncs = 0;
    
    return (code);
}



static int SCE_Media_LocatelFuncIDExt(const char *ext)
{
    /* compteurs:
       - i: parcour des structures de fonction
       - j: parcour de la chaine des extensions de la structure */
    unsigned int j;
    int i;
    
    /* booleen: = 0 si une incoherence est detectee lors de la comparaison */
    int yaisse=0;
    
    /* attention a bien noter la difference entre les deux :
       - extlen: longueur de l'extension
       - extSlen: longueur des extensions de la
                  fonction qui est en cours de scan */
    size_t extlen, extslen;
    
    /* pointeur sur la chaine renvoyee par strstr */
    char *ptrstr = NULL;
    
    extlen = strlen(ext);
    
    for(i=0; i<n_lFuncs; i++)
    {
        extslen = strlen(l_Func[i].ext);
        
        /* si toutes les extensions reunies de la structure
           ont une taille inferieur a l'extension demandee
           c'est meme pas la peine de chercher :/ */
        if(extslen < extlen)
            continue;
        
        for(j=0; j<extslen; j++)
        {
            ptrstr = strstr(&l_Func[i].ext[j], ext);
            
            if(ptrstr != NULL)
            {
                /* la fin doit etre une fin 'normale' (espace ou fin de chaine)
                   pas un caractere courant */
                if(   ptrstr[extlen] == ' '
                   || ptrstr[extlen] == '\0')
                {
                    yaisse = 1;
                }
                /* si la chaine trouvee se trouve a l'interieur */
                if(ptrstr > l_Func[i].ext)
                {
                    /* le debut doit etre une zone vide ' '
                       signalant une separation entre deux extensions */
                    if(ptrstr[-1] == ' ')
                        yaisse = 1;
                    else
                        yaisse = 0;
                }
                
                j += (ptrstr - l_Func[i].ext);
            }
            else
                break;
        }
        
        if(yaisse)
            return (i);
    }
    
    return (-1);
}

static int SCE_Media_LocatelFuncIDMagic(const int magic)
{
    int i;
    
    for(i=0; i<n_lFuncs; i++)
    {
        if(l_Func[i].magic == magic)
            return (i);
    }
    
    return (-1);
}

static int SCE_Media_LocatesFuncIDType(const int type_id)
{
    int i;
    
    for(i=0; i<n_sFuncs; i++)
    {
        if(s_Func[i].type_id == type_id)
            return (i);
    }
    
    return (-1);
}



int SCE_Media_GenTypeID(void)
{
    static int id = 0;
    id++;
    return (id);
}


void* SCE_Media_LoadFromFile(const char *filename, void *param, int *type_id)
{
    FILE *file = NULL;
    void *media = NULL;
    const char *ext;
    int magic = 0;
    int func_id = -1;
    
    /* on considere toujours que le fichier est un fichier binaire */
    file = fopen(filename, "rb");
    if(file == NULL)
    {
        Logger_Log(SCE_FILE_NOT_FOUND);
        Logger_LogMsg("file '%s' not found", filename);
        return (NULL);
    }
    
    fread(&magic, sizeof(magic), 1, file);
    rewind(file);
    
    ext = SCE_String_GetExt((char*)filename);
    
    /* 0 n'est pas un nombre magique (hugh)
       0 represente la constante NULL lors de la registration d'un loader */
    if(magic != 0)
        func_id = SCE_Media_LocatelFuncIDMagic(magic);
    else
        func_id = -1;
    
    if(func_id > -1)
    {
        media = l_Func[func_id].load(file, filename, param);
        if(type_id != NULL)
            *type_id = l_Func[func_id].type_id;
    }
    else
    {
        if(ext != NULL)
            func_id = SCE_Media_LocatelFuncIDExt(ext);
        else    /* un fichier sans nombre magique ni extension... */
            func_id = -1;
        
        if(func_id > -1)
        {
            media = l_Func[func_id].load(file, filename, param);
            if(type_id != NULL)
                *type_id = l_Func[func_id].type_id;
        }
        else
        {
            /* la fonction de chargement du type du fichier
               n'a pas ete trouvee... */
            Logger_Log(SCE_INVALID_ARG);
            Logger_LogMsg("demande de chargement d'un type de fichier inconnu");
            
            media = NULL;
            if(type_id != NULL)
                *type_id = SCE_UNKNOW_TYPE;
        }
    }
    
    fclose(file);
    
    return (media);
}

int SCE_Media_RegisterLoader(
  int type_id,          /* type de la ressource chargee par la fonction */
  int magic,            /* nombre magique du type de fichier charge */
  const char *ext,      /* extention du type de fichier charge */
  /* fonction de chargement du media */
  void* (*fct)(FILE*, const char*, void*))
{
    size_t extlen;
    
    if(fct == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (-1);
    }
    
    n_lFuncs++;
    
    l_Func = SCE_realloc(l_Func, n_lFuncs*sizeof(*l_Func));
    if(l_Func == NULL)
    {
        Logger_LogSrc();
        n_lFuncs--;
        return (-1);
    }
    
    l_Func[n_lFuncs-1].type_id = type_id;
    l_Func[n_lFuncs-1].magic = magic;
    l_Func[n_lFuncs-1].ext = NULL;
    
    if(ext != NULL)
    {
        extlen = strlen(ext);
        
        l_Func[n_lFuncs-1].ext = SCE_malloc(sizeof(char)*(extlen+1));
        if(l_Func[n_lFuncs-1].ext == NULL)
        {
            Logger_LogSrc();
            return (-1);
        }
        
        strcpy(l_Func[n_lFuncs-1].ext, ext);
        
        /* l_Func[n_lFuncs-1].ext[extlen] = '\0'; */
    }
    
    l_Func[n_lFuncs-1].load = fct;
    
    return (0);
}



int SCE_Media_SaveToFile(void* param, const char *filename, int type_id)
{
    int func_id = -1;
    
    func_id = SCE_Media_LocatesFuncIDType(type_id);
    if(func_id == -1)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (-1);
    }
    
    return (s_Func[func_id].save(filename, param));
}

int SCE_Media_RegisterSaver(int type_id, int (*fct)(const char*, void*))
{
    n_sFuncs++;
    
    s_Func = SCE_realloc(s_Func, n_sFuncs*sizeof(*s_Func));
    if(s_Func == NULL)
    {
        Logger_LogSrc();
        n_sFuncs--;
        return (-1);
    }
    
    s_Func[n_sFuncs-1].type_id = type_id;
    s_Func[n_sFuncs-1].save = fct;
    
    return (0);
}
